﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows;
using System.Windows.Threading;

namespace Repic
{
	// base printer class, basically exists so that we can hook up a virtual printer to a real printer and get
	// a real-time visualization of a hardware print
	public class BasePrinter : IPrinter
	{
		public Window Parent { get; set; }
		public IPrinter Viewer { get; set; }
		public EventHandler Ready { get; set; }

		public BasePrinter(Window parent)
		{
			this.Parent = parent;

			// kick off a worker thread to do the rendering
			BackgroundWorker worker = new BackgroundWorker();
			worker.DoWork += new DoWorkEventHandler(ThreadProc);
			worker.RunWorkerAsync();
		}

		void ThreadProc(object sender, DoWorkEventArgs e)
		{
			try
			{
				while (true)
				{
					// Sleep for a bit so we don't tie up 100% of the CPU
					Thread.Sleep(1000 / 100);

					// do our work
					Update();

					// tell anyone watching us that we're ready to receive data
					if (!this.IsBusy && (this.Ready != null))
						this.Ready(this, new EventArgs());
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show("Caught exeption in print thread, application will now exit: " + ex.ToString(),
					"Critical Error!", MessageBoxButton.OK, MessageBoxImage.Error);
				this.Parent.Dispatcher.Invoke(DispatcherPriority.Normal, new Action(delegate()
				{
					Application.Current.Shutdown(0);
				}));
			}
		}

		protected virtual void Update()
		{
		}

		public virtual void Reset()
		{
		}

		public virtual bool IsBusy
		{
			get
			{
				if (this.Viewer != null)
					return this.Viewer.IsBusy;
				return false;
			}
		}

		// print commands

		public virtual void GoHome(PrintJob job, double? X, double? Y, double? Z)
		{
			if (this.Viewer != null)
				this.Viewer.GoHome(job, X, Y, Z);
		}

		public virtual void ControlledMove(PrintJob job, ExtrudeData extrudeData)
		{
			if (this.Viewer != null)
				this.Viewer.ControlledMove(job, extrudeData);
		}

		public virtual void Dwell(PrintJob job, double dwellTime)
		{
			if (this.Viewer != null)
				this.Viewer.Dwell(job, dwellTime);
		}
	}
}
