﻿/*

FiveD Implementation Notes
==========================
 
The math behind FiveD extrusion is the same as that for accelerated motion, where 'feed rate' is
analogous to 'velocity'. From newtonian physics we have the basic equation of accelerated motion:

1) v=u+a*t

where	v = final velocity (end feedrate)
		u = initial velocity (start feedrate)
		a = acceleration
		t = time

By integrating with respect to t we can calculate total movement/extrusion distance over a given time
period:

2) s=u*t + 0.5*a*t^2

This is the equation we will use at run-time to determine how far the stepper motors should move. First,
we must calculate 'a', which we can do if we know the total time for the entire extrusion. By rearranging
1) to solve for 'a' and substituting into 2) we arrive at equation 3) which allows us to calculate total
extrude time 't' given start and end feed rates and the total distance 's', which is how extrusion is
specified in gcode:

3) t = 2 * s / (u + v)

(Note that conceptually this makes sense, it's simply distance divided by the average feed rate).

In the case of RepRap our feed rates are specified in mm/minute, so the calculated 't' value is converted
from minutes to the interrupt tick-rate on the micro-controller (i.e. 10000/second in the case of Repic).
This value is then rounded to the nearest integer and equation 1) is used to calculate 'a'. The rounding
of 't' to the nearest integer often results in a small error being introduced to 'a' which in turn can
cause the target feed rate value 'v' to be different at the end of the extrude. This error is well within
the tolerences of RepRap though (typically < 0.01%) and the rounding ensures that the final iterated values
of movement and extrusion equal the target 's' values exactly, which is more important.

To do the interpolation we convert to stepper values by multiplying distance by the step rate of the
stepper motor in question (typically ~80 for the X and Y axis on Mendel). Each step requires two
interrupts, one to set the driver signal to 1 and another to set it back to 0, so we double this value
and then scale equation 2) by the result in order to calculate the number of step signals generated for
any given t. We do not, however, need to fully evaluate equation 2) for each iteration in the main loop
because s(t+1) can be calculated as a function of s(t):

	s(t+1)	= u * (t+1) + 0.5 * a * (t+1) * (t+1)
			= u * t + u + 0.5 * a * (t * t + 2 * t + 1)
			= (u * t + 0.5 * a * t * t) + a * t + 0.5 * a + u 
			= s(t) + (a * t) + (0.5 * a + u);
 
The algorithm can therefore be implemented iteratively with floating-point variables using addition only
in the inner loop:
 
		dNumSteps = 0.5 * a + u
		ddNumSteps = a
		numSteps = 0
		for (i=0; i<t; i++, numSteps += dNumSteps, dNumSteps += ddNumSteps)
			... use numSteps here...

In order to implement this code on a 16-bit PIC micro-controller the values are stored in 32-bit integer
variables as 1:31 fixed-point values. Only 1 integer bit is needed for dNumSteps and ddNumSteps because
RepRap cannot generate more than 1 step per interrupt tick, i.e. both values are always <= 1. We do lose
all integer bits in numSteps apart from the LSB, but that's ok because only the LSB is needed to determine
when the signal sent to the stepper driver should be toggled.

Unfortunately, even 31 bits of precision can cause enough loss of precision over the course of an extrude
to result in the final value of numSteps being slightly off-target. The total amount of this accumulated
error at any given moment can be calculated exactly and is equal to the fractional component of ddNumSteps
lost after conversion to 1:31 multiplied by the number of iterations:

	ddNumStepsScaled = ddNumSteps * (1 << 31);		// scale up for 1:31 fixed-point
	ddNumStepsInt = (int)ddNumStepsScaled;			// and store in a 32-bit integer
	errorTerm = (ddNumStepsScaled - ddNumStepsInt);	// the difference is the error for each iteration
	totalError = errorTerm * t;						// total error for the entire extrude

 We can adjust for this by calculating this total error term iteratively using Brensenham's line algorithm
and adding the integer component of the result to numSteps during the inner loop in addition to dNumSteps.
This results in perfect FiveD integrated extrusion with zero accumulated error over the ranges of distance,
feed rate and acceleration typically used in RepRap.

*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

namespace Repic
{
	public class FiveD
	{
		public static void Test(ExtrudeData extrudeData, Vector5D stepRates)
		{
			if ((extrudeData.Start.F == 0) && (extrudeData.End.F == 0))
				return; // nothing to do

			const double EPSILON = 0.01;	// 1/10th resolution of the machine
			//const double EPSILON_SQUARED = EPSILON * EPSILON;

			// clamp the XYE feedrates to the limit of the machine
			//extrudeData.Start.F = Math.Min(extrudeData.Start.F, 3500);
			//extrudeData.End.F = Math.Min(extrudeData.End.F, 3500);

			// clamp the XYE feedrates to the limit of the machine
			if ((extrudeData.Start.X != extrudeData.End.X) || (extrudeData.Start.Y != extrudeData.End.Y))
			{
				double maxFeedrateX = RepRapPrinter.TICKS_PER_MINUTE / RepRapPrinter.X_STEPS_PER_MM / 2;
				extrudeData.Start.F = Math.Min(extrudeData.Start.F, maxFeedrateX);
				extrudeData.End.F = Math.Min(extrudeData.End.F, maxFeedrateX);
				Debug.Assert((extrudeData.Start.F != 0) || (extrudeData.End.F != 0));
				double maxFeedrateY = RepRapPrinter.TICKS_PER_MINUTE / RepRapPrinter.Y_STEPS_PER_MM / 2;
				extrudeData.Start.F = Math.Min(extrudeData.Start.F, maxFeedrateY);
				extrudeData.End.F = Math.Min(extrudeData.End.F, maxFeedrateY);
			}
			else if (extrudeData.Start.E != extrudeData.End.E)
			{
				double maxFeedrateE = RepRapPrinter.TICKS_PER_MINUTE / RepRapPrinter.E_STEPS_PER_MM / 2;
				extrudeData.Start.F = Math.Min(extrudeData.Start.F, maxFeedrateE);
				extrudeData.End.F = Math.Min(extrudeData.End.F, maxFeedrateE);
				Debug.Assert((extrudeData.Start.F != 0) || (extrudeData.End.F != 0));
			}

			// calculate how far we're going to move along each axis
			Vector5D delta = extrudeData.End - extrudeData.Start;
			delta.X = Math.Abs(delta.X);
			delta.Y = Math.Abs(delta.Y);
			delta.E = Math.Abs(delta.E);

			// calculate how far we're going to move the extrude head
			double distance = Math.Max(delta.X, delta.Y);

			// if it's 0 then the only thing changing is e
			if (distance < EPSILON)
				distance = delta.E;

			// if it's still zero then the only thing changing is the feed rate
			if (distance < EPSILON)
				return;

			// solve for t
			double fi = extrudeData.Start.F;
			double fe = extrudeData.End.F;
			double t = 2 * distance / (fi + fe);

			// convert t from min to interrupts (10,000 per second) and round to nearest integer
			int ticks = (int)Math.Round(t * stepRates.F, MidpointRounding.AwayFromZero);

			// calculate a modified value of a
			double ascale = stepRates.F / ticks;
			double a = 2.0 * (distance * ascale - fi) * ascale;

			delta.X = Math.Round(delta.X * stepRates.X * 2, MidpointRounding.AwayFromZero);
			delta.Y = Math.Round(delta.Y * stepRates.Y * 2, MidpointRounding.AwayFromZero);
			delta.E = Math.Round(delta.E * stepRates.E * 2, MidpointRounding.AwayFromZero);
			Debug.Assert(delta.X <= ticks);
			Debug.Assert(delta.Y <= ticks);
			Debug.Assert(delta.E <= ticks);

			double fpscale = 1U << 31;
			int fpscalei = (int)(0.5 * fpscale);
			double tscale = t / ticks;

			double ax = a * delta.X * tscale * tscale / distance,
					fix = fi * delta.X * tscale / distance,
					dx = (ax * 0.5 + fix) * fpscale,
					ddx = ax * fpscale;
			int dxi = (int)Math.Floor(dx),
				ddxi = (int)Math.Floor(ddx),
				xi = 0,
				x = 0x40000000,
				slopeX = (int)(ticks * (ddx - ddxi)),
				errorX = (int)((dx - dxi - 1) * ticks);

			double ay = a * delta.Y * tscale * tscale / distance,
					fiy = fi * delta.Y * tscale / distance,
					dy = (ay * 0.5 + fiy) * fpscale,
					ddy = ay * fpscale;
			int dyi = (int)Math.Floor(dy),
				ddyi = (int)Math.Floor(ddy),
				yi = 0,
				y = 0x40000000,
				slopeY = (int)(ticks * (ddy - ddyi)),
				errorY = (int)((dy - dyi - 1) * ticks);

			double ae = a * delta.E * tscale * tscale / distance,
					fie = fi * delta.E * tscale / distance,
					de = (ae * 0.5 + fie) * fpscale,
					dde = ae * fpscale;
			int dei = (int)Math.Floor(de),
				ddei = (int)Math.Floor(dde),
				ei = 0,
				e = 0x40000000,
				slopeE = (int)(ticks * (dde - ddei)),
				errorE = (int)((de - dei - 1) * ticks);

			for (int tickNum = 0; tickNum <= ticks; tickNum++)
			{
				//x = fix * tickNum + 0.5 * ax * tickNum * tickNum;

				if ((x & 0x80000000) != 0)
				{
					x &= 0x7fffffff;
					xi++;
				}
				x += dxi;
				dxi += ddxi;
				errorX += slopeX;
				if (errorX >= 0)
				{
					errorX -= ticks;
					dxi++;
				}

				if ((y & 0x80000000) != 0)
				{
					y &= 0x7fffffff;
					yi++;
				}
				y += dyi;
				dyi += ddyi;
				errorY += slopeY;
				if (errorY >= 0)
				{
					errorY -= ticks;
					dyi++;
				}

				if ((e & 0x80000000) != 0)
				{
					e &= 0x7fffffff;
					ei++;
				}
				e += dei;
				dei += ddei;
				errorE += slopeE;
				if (errorE >= 0)
				{
					errorE -= ticks;
					dei++;
				}
			}
			double xt = fix * ticks + 0.5 * ax * ticks * ticks;
			double yt = fiy * ticks + 0.5 * ay * ticks * ticks;
			double et = fie * ticks + 0.5 * ae * ticks * ticks;
			Debug.Assert(Math.Abs(xi - delta.X) <= 0);
			Debug.Assert(Math.Abs(xt - delta.X) <= 0.5);
			Debug.Assert(Math.Abs(yi - delta.Y) <= 0);
			Debug.Assert(Math.Abs(yt - delta.Y) <= 0.5);
			Debug.Assert(Math.Abs(ei - delta.E) <= 0);
			Debug.Assert(Math.Abs(et - delta.E) <= 0.5);
		}
	}
}
