﻿using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Media.Media3D;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.ComponentModel;
using System.Threading;
using System.Windows.Threading;
using System.Runtime.InteropServices;
using System.Diagnostics;
using System.Reflection;
using _3DTools;

namespace Repic
{
	/// <summary>
	/// Interaction logic for MainWindow.xaml
	/// </summary>
	public partial class MainWindow : Window, INotifyPropertyChanged
	{
		// needed for two-way binding
		public event PropertyChangedEventHandler PropertyChanged;
		public void OnPropertyChanged(PropertyChangedEventArgs e)
		{
			if (PropertyChanged != null)
				PropertyChanged(this, e);
		}

		private int _ExtruderTemperature = 7;
		public int ExtruderTemperature
		{
			get { return this._ExtruderTemperature; }
			set
			{
				this._ExtruderTemperature = value;
				OnPropertyChanged(new PropertyChangedEventArgs("ExtruderTemperature"));
			}
		}

		// this document object hold the thing currently being printed
		RepRapDocument Document = null;

		// these are the printer, one for virtual mendel mode and another for the target hardware
		public VirtualRepRap VirtualRepRap { get; set; }
		public RepRapPrinter RepRap { get; set; }
		public PrintJob CurrentJob { get; set; }

		private RepicMode _CurrentMode = RepicMode.VirtualRepRap;
		public RepicMode CurrentMode
		{
			get { return this._CurrentMode; }
			set
			{
				if (this._CurrentMode != value)
				{
					if (this.CurrentJob != null)
					{
						if (value == RepicMode.VirtualRepRap)
							this.CurrentJob.Printer = this.VirtualRepRap;
						else if (value == RepicMode.HardwarePrint)
						{
							this.CurrentJob.PrintState = PrintState.Stopped;
							this.VirtualRepRap.Reset();
							this.RepRap.Reset();
							this.CurrentJob.Printer = this.RepRap;
						}
					}
					this._CurrentMode = value;
					OnPropertyChanged(new PropertyChangedEventArgs("CurrentMode"));
				}
			}
		}

		public bool ExtruderHeater
		{
			get { return (this.RepRap!=null) ? this.RepRap.ExtruderHeater : false; }
			set
			{
				if (this.RepRap != null)
					this.RepRap.ExtruderHeater = value;
				OnPropertyChanged(new PropertyChangedEventArgs("ExtruderHeater"));
			}
		}

		public bool HeatbedHeater
		{
			get { return (this.RepRap != null) ? this.RepRap.HeatbedHeater : false; }
			set
			{
				if (this.RepRap != null)
					this.RepRap.HeatbedHeater = value;
				OnPropertyChanged(new PropertyChangedEventArgs("HeatbedHeater"));
			}
		}

		private void OnVirtualRepRap(object sender, RoutedEventArgs e)
		{
			this.CurrentMode = RepicMode.VirtualRepRap;
		}

		private void OnHardwareTest(object sender, RoutedEventArgs e)
		{
			this.CurrentMode = RepicMode.HardwareTest;
		}

		private void OnHardwarePrint(object sender, RoutedEventArgs e)
		{
			this.CurrentMode = RepicMode.HardwarePrint;
		}

		public MainWindow()
		{
			string name = Assembly.GetExecutingAssembly().GetName().Name.ToString();
			string major = Assembly.GetExecutingAssembly().GetName().Version.Major.ToString();
			string minor = Assembly.GetExecutingAssembly().GetName().Version.Minor.ToString();
			this.Title = String.Format("{0} - Beta {1}.{2}", name, major, minor);

			InitializeComponent();
			this.DataContext = this;
			this.CurrentMode = RepicMode.VirtualRepRap;			
		}

		private void OnClosing(object sender, CancelEventArgs e)
		{
			if (this.CurrentJob != null)
			{
				this.CurrentJob.Dispose();
				this.CurrentJob = null;
			}

			SaveSettings();
		}

		private void OnLoaded(object sender, RoutedEventArgs e)
		{
			// create and intialize the printers
			this.Document = new RepRapDocument();
			this.VirtualRepRap = new VirtualRepRap(this);
			this.RepRap = new RepRapPrinter(this);
			this.RepRap.Viewer = this.VirtualRepRap;

			// we want notification whenever the USB status changes
			this.RepRap.OnStatusChanged += this.OnUsbStatusChanged;
			this.ExtruderThermometer.DataContext = this.RepRap;
			this.XEndStopControl.DataContext = this.RepRap;
			this.YEndStopControl.DataContext = this.RepRap;
			this.ZEndStopControl.DataContext = this.RepRap;

			LoadSettings();
		}

		// todo: clean this up with proper bindings etc
		private void LoadSettings()
		{
			Repic.Properties.Settings.Default.Reload();
			RepRapPrinter.X_STEPS_PER_MM = Repic.Properties.Settings.Default.X_STEPS_PER_MM;
			RepRapPrinter.Y_STEPS_PER_MM = Repic.Properties.Settings.Default.Y_STEPS_PER_MM;
			RepRapPrinter.Z_STEPS_PER_MM = Repic.Properties.Settings.Default.Z_STEPS_PER_MM;
			RepRapPrinter.E_STEPS_PER_MM = Repic.Properties.Settings.Default.E_STEPS_PER_MM;
		}

		private void SaveSettings()
		{
			
			Repic.Properties.Settings.Default.X_STEPS_PER_MM = RepRapPrinter.X_STEPS_PER_MM;
			Repic.Properties.Settings.Default.Y_STEPS_PER_MM = RepRapPrinter.Y_STEPS_PER_MM;
			Repic.Properties.Settings.Default.Z_STEPS_PER_MM = RepRapPrinter.Z_STEPS_PER_MM;
			Repic.Properties.Settings.Default.E_STEPS_PER_MM = RepRapPrinter.E_STEPS_PER_MM;
			Repic.Properties.Settings.Default.Save();
		}

		private void OnUsbStatusChanged(Usb sender, ConnectionStatus oldStatus, ConnectionStatus newStatus)
		{
			this.Dispatcher.Invoke(DispatcherPriority.Normal, new Action(delegate()
			{
				switch (newStatus)
				{
					case ConnectionStatus.Unknown:
						this.Status.Text = "Status: Unknown";
						break;

					case ConnectionStatus.Disconnected:
						this.Status.Text = "Status: Disconnected";
						break;

					case ConnectionStatus.Broken:
						this.Status.Text = "Status: Broken";
						break;

					case ConnectionStatus.Connected:
						this.Status.Text = "Status: Connected";
						break;
				}
			}
			));
		}

		private void OnLoad(object sender, RoutedEventArgs e)
		{	
			this.CurrentFile.Text = "";

			// Create OpenFileDialog
			Microsoft.Win32.OpenFileDialog dlg = new Microsoft.Win32.OpenFileDialog();
 
			// Set filter for file extension and default file extension
			dlg.DefaultExt = ".txt";
			dlg.Filter = "GCode file documents (.gcode)|*.gcode";
 
			// Display OpenFileDialog by calling ShowDialog method
			Nullable<bool> result = dlg.ShowDialog();
 
			// Get the selected file name and display in a TextBox
			if (result != true)
				return;

			using (new CursorHelper(this, Cursors.Wait))
			{
				LoadFile(dlg.FileName);
				this.CurrentFile.Text = dlg.FileName;
			}

			if (this.CurrentJob != null)
			{
				this.CurrentJob.PrintState = PrintState.Stopped;
				this.CurrentJob = null;
			}
		}

		private void CommandBinding_Executed(object sender, ExecutedRoutedEventArgs e)
		{
			this.Close();
		}

		private void LoadFile(string filename)
		{
			// load the file
			try
			{
				this.Document.Load(filename);
			}
			catch (Exception ex)
			{
				MessageBox.Show("Error loading gcode file: " + ex.Message);
				return;
			}

			this.VirtualRepRap.Reset();
		}
		

		private void XAxisLeftPressedDown(object sender, MouseButtonEventArgs e) { this.RepRap.XStepper = -1; }
		private void XAxisLeftPressedUp(object sender, MouseButtonEventArgs e) { this.RepRap.XStepper = 0; }
		private void XAxisRightPressedDown(object sender, MouseButtonEventArgs e) { this.RepRap.XStepper = 1; }
		private void XAxisRightPressedUp(object sender, MouseButtonEventArgs e) { this.RepRap.XStepper = 0; }
		private void XAxisHome(object sender, RoutedEventArgs e) { this.RepRap.GoHome(null, 0.0, null, null); }

		private void YAxisBackPressedDown(object sender, MouseButtonEventArgs e) { this.RepRap.YStepper = -1; }
		private void YAxisBackPressedUp(object sender, MouseButtonEventArgs e) { this.RepRap.YStepper = 0; }
		private void YAxisForwardPressedDown(object sender, MouseButtonEventArgs e) { this.RepRap.YStepper = 1; }
		private void YAxisForwardPressedUp(object sender, MouseButtonEventArgs e) { this.RepRap.YStepper = 0; }
		private void YAxisHome(object sender, RoutedEventArgs e) { this.RepRap.GoHome(null, null, 0.0, null); }

		private void ZAxisDownPressedDown(object sender, MouseButtonEventArgs e) { this.RepRap.ZStepper = -1; }
		private void ZAxisDownPressedUp(object sender, MouseButtonEventArgs e) { this.RepRap.ZStepper = 0; }
		private void ZAxisUpPressedDown(object sender, MouseButtonEventArgs e) { this.RepRap.ZStepper = 1; }
		private void ZAxisUpPressedUp(object sender, MouseButtonEventArgs e) { this.RepRap.ZStepper = 0; }
		private void ZAxisHome(object sender, RoutedEventArgs e) { this.RepRap.GoHome(null, null, null, 0.0); }

		private void EAxisReversePressedDown(object sender, MouseButtonEventArgs e) { this.RepRap.EStepper = -1; }
		private void EAxisReversePressedUp(object sender, MouseButtonEventArgs e) { this.RepRap.EStepper = 0; }
		private void EAxisForwardPressedDown(object sender, MouseButtonEventArgs e) { this.RepRap.EStepper = 1; }
		private void EAxisForwardPressedUp(object sender, MouseButtonEventArgs e) { this.RepRap.EStepper = 0; }

		private void OnAbout(object sender, RoutedEventArgs e)
		{
			new AboutBox().ShowDialog();
		}

		private void OnDeviceSetUp(object sender, RoutedEventArgs e)
		{
			// if we're not connected then we can't do anything (actually let it go for now, because settings are stored locally)
			if (this.RepRap.ConnectionStatus != ConnectionStatus.Connected)
			{
				//MessageBox.Show("No RepRap device connected.");
				//return;
			}

			try
			{
				// show the dialog box
				SetUpDlg dlg = new SetUpDlg();
				dlg.Owner = this;
				dlg.XStepsPerMM = RepRapPrinter.X_STEPS_PER_MM.ToString("0.00");
				dlg.YStepsPerMM = RepRapPrinter.Y_STEPS_PER_MM.ToString("0.00");
				dlg.ZStepsPerMM = RepRapPrinter.Z_STEPS_PER_MM.ToString("0.00");
				dlg.EStepsPerMM = RepRapPrinter.E_STEPS_PER_MM.ToString("0.00");
				dlg.ShowDialog();
				if (dlg.DialogResult.HasValue && dlg.DialogResult.Value)
				{
					RepRapPrinter.X_STEPS_PER_MM = Convert.ToDouble(dlg.XStepsPerMM);
					RepRapPrinter.Y_STEPS_PER_MM = Convert.ToDouble(dlg.YStepsPerMM);
					RepRapPrinter.Z_STEPS_PER_MM = Convert.ToDouble(dlg.ZStepsPerMM);
					RepRapPrinter.E_STEPS_PER_MM = Convert.ToDouble(dlg.EStepsPerMM);
					SaveSettings();
					return;
				}
			}
			catch
			{
				if (this.RepRap.ConnectionStatus != ConnectionStatus.Connected)
					MessageBox.Show("RepRap device no longer connected, changes discarded.");
				else
					MessageBox.Show("Error occurred sending settings to RepRap device, changes discarded.");
			}
		}

		private void OnDeviceSettings(object sender, RoutedEventArgs e)
		{
			// if we're not connected then we can't do anything
			if (this.RepRap.ConnectionStatus != ConnectionStatus.Connected)
			{
			    MessageBox.Show("No RepRap device connected.");
			    return;
			}

			try
			{
				// show the dialog box
				SettingsDlg dlg = new SettingsDlg();
				dlg.ExtruderTemp = this.RepRap.GetTargetTemp();
				dlg.Owner = this;
				if (dlg.ShowDialog().Value)
					return;
				this.RepRap.SetTargetTemp(dlg.ExtruderTemp);
			}
			catch
			{
				if (this.RepRap.ConnectionStatus != ConnectionStatus.Connected)
					MessageBox.Show("RepRap device no longer connected, changes discarded.");
				else
					MessageBox.Show("Error occurred sending settings to RepRap device, changes discarded.");
			}
		}

		private void OnPrint(object sender, RoutedEventArgs e)
		{
			// make sure we've loaded a document
			if (!this.Document.Loaded)
			{
				MessageBox.Show("Nothing to print!");
				return;
			}

			// if we're in hardware test mode then prompt to switch to print mode
			if (this.CurrentMode == RepicMode.HardwareTest)
			{
				if (MessageBox.Show("Switch to Hardware Print mode and start printing?",
					"Cannot print in Test mode", MessageBoxButton.YesNo, MessageBoxImage.Question) == MessageBoxResult.No)
					return;
				this.CurrentMode = RepicMode.HardwarePrint;
			}
			
			// are we in hardware print mode?
			if (this.CurrentMode == RepicMode.HardwarePrint)
			{
				// yes, so make sure we're connected
				if (this.RepRap.ConnectionStatus != ConnectionStatus.Connected)
				{
					MessageBox.Show("No RepRap hardware connected.", "Error", MessageBoxButton.OK, MessageBoxImage.Hand);
					return;
				}
				
				// and make sure the heater is connected
				if (!this.ExtruderHeater)
				{
					MessageBoxResult result = MessageBox.Show("Extruder heater is off, turn it on first?", "Warning", MessageBoxButton.YesNoCancel, MessageBoxImage.Warning);
					switch (result)
					{
						case MessageBoxResult.Yes:
							this.ExtruderHeater = true;
							break;

						case MessageBoxResult.Cancel:
							return;
					}
				}
			}

			// if we don't have an active print job then create one and start printing
			if (this.CurrentJob == null)
			{
				// select the printer to use based on the current mode
				this.VirtualRepRap.Reset();
				this.RepRap.Reset();
				IPrinter printer;
				if (this.CurrentMode == RepicMode.VirtualRepRap)
					printer = this.VirtualRepRap;
				else if (this.CurrentMode == RepicMode.HardwarePrint)
				{
					printer = this.RepRap;
				}
				else
				{
					// shouldn't ever reach here
					MessageBox.Show("Don't know how to print in current mode.", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
					return;
				}

				// create the print job
				this.CurrentJob = new PrintJob(this, this.Document);
				this.CurrentJob.OnComment += OnComment;
				this.CurrentJob.Printer = printer;
			}

			// if we're stopped then start printing
			if (this.CurrentJob.PrintState == PrintState.Stopped)
			{
				this.VirtualRepRap.Reset();
				this.RepRap.Reset();
				this.CurrentJob.PrintState = PrintState.Printing;
				this.btnStop.IsEnabled = true;
			}

			// if we're printing then pause
			else if (this.CurrentJob.PrintState == PrintState.Printing)
			{
				this.CurrentJob.PrintState = PrintState.Paused;
				this.btnStop.IsEnabled = true;
			}

			// if we're paused then resume
			else if (this.CurrentJob.PrintState == PrintState.Paused)
			{
				this.CurrentJob.PrintState = PrintState.Printing;
				this.btnStop.IsEnabled = true;
			}
		}

		private void OnStop(object sender, RoutedEventArgs e)
		{
			this.btnStop.IsEnabled = false;
			if (this.CurrentJob != null)
				this.CurrentJob.PrintState = PrintState.Stopped;
		}

		private void OnComment(object sender, string comment)
		{
			this.Dispatcher.Invoke(DispatcherPriority.Normal, new Action(delegate()
			{
				this.Status.Text = comment;
			}));
		}

	}
	
}
