﻿// represents the printing of a 3d model. the actual data for the model is stored in ModelDocument, and the IPrinter
// classes (VirtualPrinter and RepRapPrinter) are responsible for handling communication with the target device. this
// class represents the job itself, i.e. the position in the print queue and current state of the print

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows;
using System.Windows.Threading;
using System.ComponentModel;
using System.Diagnostics;
using System.Text.RegularExpressions;

namespace Repic
{
	public delegate void CommentHandler(object sender, string comment);

	public class PrintJob : IDisposable
	{
		public CommentHandler OnComment { get; set; }

		public Window Parent;
		public RepRapDocument Document;
		
		private IPrinter _Printer = null;
		public IPrinter Printer
		{
			get
			{
				return this._Printer;
			}
			set
			{
				if (this._Printer != null)
					this._Printer.Ready -= this.OnPrinterReady;
				this._Printer = value;
				if (this._Printer != null)
					this._Printer.Ready += this.OnPrinterReady;
			}
		}

		int PrintPosition = 0;
		//PositioningType CurrentPositioningType = PositioningType.Absolute;
		Vector5D CurrentPosition = new Vector5D(0.0, 0.0, 0.0, 0.0, 0.0);
		Vector5D NextPosition = new Vector5D(0.0, 0.0, 0.0, 0.0, 0.0);
		Vector5D OffsetPosition = new Vector5D(0.0, 0.0, 0.0, 0.0, 0.0);
	
		private delegate void PrintCommandHandler(string args);
		private Dictionary<int, PrintCommandHandler> GCodeHandlers = new Dictionary<int, PrintCommandHandler>();
		private Dictionary<int, PrintCommandHandler> MCodeHandlers = new Dictionary<int, PrintCommandHandler>();

		public UnitType CurrentUnitType = UnitType.Metric;
		double PwmExtruderSpeed = 0.0;
		bool PwmExtruderEnabled = false;

		public bool Printing { get { return this.PrintState == PrintState.Printing; } }

		private volatile PrintState _PrintState = PrintState.Stopped;
		public PrintState PrintState
		{
			get
			{
				return this._PrintState;
			}
			set
			{
				if (value == this._PrintState)
					return;

				// if we're starting or resuming a print then we need to resume the print thread
				if (value==PrintState.Printing)
				{
					// if we were at the end of the document then we need to restart the print job
					if ((this._PrintState == PrintState.Stopped) || this.EndOfDocument)
						this.Reset();
					
					// starting a print, kick-start the thread
					this._PrintState = value;
				}

				// if we were printing then we need to destroy the print thread
				else if (this._PrintState == PrintState.Printing)
				{
					this._PrintState = value;
				}

			}
		}

		public volatile bool Closing = false;

		public PrintJob(Window parent, RepRapDocument document)
		{
			this.Parent = parent;
			this.Document = document;
			SetUpCommandHandlers();
		}

		public void Dispose()
		{
			// stop any printing
			this.PrintState = PrintState.Stopped;
			this.Closing = true;
		}
		
		void OnPrinterReady(object sender, EventArgs args)
		{
			// if we're not printing then there's nothing to do
			if (!this.Printing)
				return;

			// if we're at the end of the print sequence then pause printing
			if (this.EndOfDocument)
			{
				this.PrintState = PrintState.Paused;
				return;
			}

			// keep printing for 1/60th of a second  (or until the queue fills up if we're printing to hardware)
			DateTime start = DateTime.Now;
			do
			{
				// make sure the printer is ready to accept more commands
				if ((this.Printer != null) && (this.Printer.IsBusy))
					break;

				// process and send the next command
				Step();
			} while (((DateTime.Now - start) < TimeSpan.FromMilliseconds(1000 / 60)) && (!this.EndOfDocument));
		}

		public bool EndOfDocument
		{
			get { return this.PrintPosition >= this.Document.GCodes.Length; }
		}

		public void Reset()
		{
			this.PrintPosition = 0;
			this.CurrentUnitType = UnitType.Metric;
			//this.CurrentPositioningType = PositioningType.Absolute;
			this.CurrentPosition = new Vector5D(0.0, 0.0, 0.0, 0.0, 0.0);
			this.NextPosition = new Vector5D(0.0, 0.0, 0.0, 0.0, 0.0);
			this.OffsetPosition = new Vector5D(0.0, 0.0, 0.0, 0.0, 0.0);

			if (this.Printer != null)
				this.Printer.Reset();
		}

		public void SetUpCommandHandlers()
		{
			// gcodes
			GCodeHandlers[0] = RapidMove;
			GCodeHandlers[1] = ControlledMove;
			GCodeHandlers[28] = GoHome;
			GCodeHandlers[4] = Dwell;
			GCodeHandlers[20] = SetImperial;
			GCodeHandlers[21] = SetMetric;
			GCodeHandlers[90] = AbsolutePositioning;
			GCodeHandlers[91] = IncrementalPositioning;
			GCodeHandlers[92] = SetPosition;

			// mcodes
			MCodeHandlers[0] = ShutDown;
			MCodeHandlers[140] = SetBedTemp;
			MCodeHandlers[113] = SetExtruderPot;
			MCodeHandlers[109] = SetExtruderTempWait;
			MCodeHandlers[107] = CoolerOff;
			MCodeHandlers[106] = CoolerOn;
			MCodeHandlers[104] = SetExtruderTemp;
			MCodeHandlers[105] = GetExtruderTemp;
			MCodeHandlers[110] = NewPrint;
			MCodeHandlers[108] = ExtruderSpeed;
			MCodeHandlers[101] = ExtruderOn;
			MCodeHandlers[103] = ExtruderOff;
			MCodeHandlers[141] = SetChamberTemp;
			MCodeHandlers[142] = SetHoldingPressure;
		}

		public void Step()
		{
			Match match;

			// if we've finished printing then don't bother
			if (this.EndOfDocument)
				return;

			// get the next instruction
			string instr = this.Document.GCodes[this.PrintPosition++];

			// check for a g-code
			match = Regex.Match(instr, @"^G(?<code>\d+)\b(?<args>.*)");
			if (match.Groups["code"].ToString() != "")
			{
				int code = Convert.ToInt32(match.Groups["code"].ToString());
				Debug.Assert(GCodeHandlers.ContainsKey(code), "Unknown G-Code: " + code.ToString());
				if (GCodeHandlers.ContainsKey(code))
				{
					string args = match.Groups["args"].ToString();
					GCodeHandlers[code](args);
				}
				return;
			}

			// check for an m-code
			match = Regex.Match(instr, @"^M(?<code>\d+)\b(?<args>.*)");
			if (match.Groups["code"].ToString() != "")
			{
				int code = Convert.ToInt32(match.Groups["code"].ToString());
				Debug.Assert(MCodeHandlers.ContainsKey(code), "Unknown M-Code: " + code.ToString());
				if (MCodeHandlers.ContainsKey(code))
				{
					string args = match.Groups["args"].ToString();
					MCodeHandlers[code](args);
				}
				return;
			}
			

			// check for a status comment
			match = Regex.Match(instr, @"^;#!(?<comment>.*)");
			if (match.Groups["comment"].ToString() != "")
			{
				string comment = match.Groups["comment"].ToString();
				if (this.OnComment != null)
					this.OnComment(this, comment);
			}
		}

		private void ParseNumberField(string args, string field, ref double val)
		{
			Match match = Regex.Match(args, field + @"(?<val>[0-9+-.]+\b)");
			if (match.Groups["val"].ToString() != "")
				val = Convert.ToDouble(match.Groups["val"].ToString());
		}

		private void ParseNumberField(string args, string field, ref double? val)
		{
			Match match = Regex.Match(args, field + @"(?<val>[0-9+-.]+\b)");
			if (match.Groups["val"].ToString() != "")
				val = Convert.ToDouble(match.Groups["val"].ToString());
		}

		/*
		 * Command Handlers
		 */

		private void RapidMove(string args)
		{
			// just do a regular controlled move for now
			ControlledMove(args);
		}

		private void ControlledMove(string args)
		{
			this.NextPosition = this.CurrentPosition;
			ParseNumberField(args, "X", ref this.NextPosition.X);
			ParseNumberField(args, "Y", ref this.NextPosition.Y);
			ParseNumberField(args, "Z", ref this.NextPosition.Z);
			ParseNumberField(args, "E", ref this.NextPosition.E);
			ParseNumberField(args, "F", ref this.NextPosition.F);

			// if we're using the deprecated method of extruder control (i.e. M108/M101/M103)
			// then calculate it here and convert to the newer method.
			if (this.PwmExtruderEnabled)
			{
				double dx = this.NextPosition.X - this.CurrentPosition.X;
				double dy = this.NextPosition.Y - this.CurrentPosition.Y;
				double dz = this.NextPosition.Z - this.CurrentPosition.Z;
				double distance = Math.Sqrt(dx * dx + dy * dy + dz * dz);

				// todo: next line is a hack, fix it! (MJF)
				//this.NextPosition.E = this.CurrentPosition.E + distance * this.PwmExtruderSpeed / this.CurrentPosition.F;
				this.NextPosition.E = this.CurrentPosition.E + distance * this.PwmExtruderSpeed / 3500;
			}

			ExtrudeData data = new ExtrudeData(this.CurrentPosition + this.OffsetPosition, this.NextPosition + this.OffsetPosition);

			// can't go past the endstops
			data.Start.X = Math.Max(data.Start.X, 0);
			data.Start.Y = Math.Max(data.Start.Y, 0);
			data.End.X = Math.Max(data.End.X, 0);
			data.End.Y = Math.Max(data.End.Y, 0);

			if (this.Printer != null)
				this.Printer.ControlledMove(this, data);
			this.CurrentPosition = this.NextPosition;
		}

		private void GoHome(string args)
		{
			double? homeX = null;
			double? homeY = null;
			double? homeZ = null;
			ParseNumberField(args, "X", ref homeX);
			ParseNumberField(args, "Y", ref homeY);
			ParseNumberField(args, "Z", ref homeZ);

			if (this.Printer != null)
				this.Printer.GoHome(this, homeX, homeY, homeZ);
			this.CurrentPosition.X = 0;
			this.CurrentPosition.Y = 0;
			this.OffsetPosition.X = 0;
			this.OffsetPosition.Y = 0;
		}

		private void Dwell(string args)
		{
			double dwellTime = 0;
			ParseNumberField(args, "P", ref dwellTime);
			if (this.Printer != null)
				this.Printer.Dwell(this, dwellTime);
		}

		private void SetImperial(string args)
		{
			this.CurrentUnitType = UnitType.Imperial;
		}

		private void SetMetric(string args)
		{
			this.CurrentUnitType = UnitType.Metric;
		}

		private void AbsolutePositioning(string args)
		{
			//this.CurrentPositioningType = PositioningType.Absolute;
		}

		private void IncrementalPositioning(string args)
		{
			// todo: implement incremental positioning
			//this.CurrentPositioningType = PositioningType.Incremental;
		}

		private void SetPosition(string args)
		{
			ParseNumberField(args, "X", ref this.CurrentPosition.X);
			ParseNumberField(args, "Y", ref this.CurrentPosition.Y);
			ParseNumberField(args, "Z", ref this.CurrentPosition.Z);
			ParseNumberField(args, "E", ref this.CurrentPosition.E);
		}

		private void ShutDown(string args)
		{
			// ignored
		}

		private void SetBedTemp(string args)
		{
			// ignored
		}

		private void SetExtruderPot(string args)
		{
			// ignored
		}

		private void SetExtruderTemp(string args)
		{
			// ignored
		}

		private void GetExtruderTemp(string args)
		{
			// ignored
		}

		private void NewPrint(string args)
		{
			// ignored
		}

		private void SetExtruderTempWait(string args)
		{
			// ignored
		}

		private void CoolerOff(string args)
		{
			// ignored
		}

		private void CoolerOn(string args)
		{
			// ignored
		}

		private void ExtruderSpeed(string args)
		{
			ParseNumberField(args, "S", ref this.PwmExtruderSpeed);
		}

		private void ExtruderOn(string args)
		{
			PwmExtruderEnabled = true;
		}

		private void ExtruderOff(string args)
		{
			PwmExtruderEnabled = false;
		}

		private void SetChamberTemp(string args)
		{
			// ignored
		}

		private void SetHoldingPressure(string args)
		{
			// ignored
		}
		
	}

	public enum UnitType
	{
		Imperial,
		Metric
	}

	public enum PositioningType
	{
		Absolute,
		Incremental
	}
}
