﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Repic
{
	public class RepRapPacket
	{
		// TODO: create a pool to manage these, it'll be faster and result in less memory fragmentation
		private byte[] _Data = null;
		public byte[] Data {get {return this._Data;}}
		
		public RepRapPacket()
		{
			this._Data = new byte[65];
			this._Data[0] = 0;
			for (int i = 1; i < this._Data.Length; i++)
				this._Data[i] = 0xff;
		}

		protected void PackInt(int pos, int val)
		{
			this.Data[pos + 0] = (byte)val;
			this.Data[pos + 1] = (byte)(val >> 8);
			this.Data[pos + 2] = (byte)(val >> 16);
			this.Data[pos + 3] = (byte)(val >> 24);
		}

		protected void PackShort(int pos, int val)
		{
			this.Data[pos + 0] = (byte)val;
			this.Data[pos + 1] = (byte)(val >> 8);
		}
	}

	public class RequestStatusPacket : RepRapPacket
	{
		public RequestStatusPacket()
		{
			this.Data[1] = (byte)MessageID.MSGID_REQUEST_STATUS;
		}
	}

	public class RequestTargetTempPacket : RepRapPacket
	{
		public RequestTargetTempPacket()
		{
			this.Data[1] = (byte)MessageID.MSGID_REQUEST_TARGET_TEMP;
		}
	}

	public class XStepperPacket : RepRapPacket
	{
		public XStepperPacket(int direction)
		{
			this.Data[1] = (byte)MessageID.MSGID_XSTEPPER;
			this.Data[2] = (byte)direction;
		}
	}

	public class YStepperPacket : RepRapPacket
	{
		public YStepperPacket(int direction)
		{
			this.Data[1] = (byte)MessageID.MSGID_YSTEPPER;
			this.Data[2] = (byte)direction;
		}
	}

	public class ZStepperPacket : RepRapPacket
	{
		public ZStepperPacket(int direction)
		{
			this.Data[1] = (byte)MessageID.MSGID_ZSTEPPER;
			this.Data[2] = (byte)direction;
		}
	}

	public class EStepperPacket : RepRapPacket
	{
		public EStepperPacket(int direction)
		{
			this.Data[1] = (byte)MessageID.MSGID_ESTEPPER;
			this.Data[2] = (byte)direction;
		}
	}

	public class ExtruderHeaterPacket : RepRapPacket
	{
		public ExtruderHeaterPacket(bool on)
		{
			this.Data[1] = (byte)MessageID.MSGID_EXTRUDER_HEATER;
			this.Data[2] = Convert.ToByte(on);
		}
	}

	public class HeatbedHeaterPacket : RepRapPacket
	{
		public HeatbedHeaterPacket(bool on)
		{
			this.Data[1] = (byte)MessageID.MSGID_HEATBED_HEATER;
			this.Data[2] = Convert.ToByte(on);
		}
	}

	public class SetTargetTempPacket : RepRapPacket
	{
		public SetTargetTempPacket(int temp, int adc)
		{
			this.Data[1] = (byte)MessageID.MSGID_SET_TARGET_TEMP;
			this.Data[2] = (byte)temp;
			PackShort(3, adc);
		}
	}

	public class GoHomePacket : RepRapPacket
	{
		public GoHomePacket(double? X, double? Y, double? Z)
		{
			this.Data[1] = (byte)MessageID.MSGID_GO_HOME;
			this.Data[2] = 0;
			if (X != null)
				this.Data[2] |= (1 << 0);
			if (Y != null)
				this.Data[2] |= (1 << 1);
			if (Z != null)
				this.Data[2] |= (1 << 2);
		}
	}

	public class ControlledMove4DPacket : RepRapPacket
	{
		public ControlledMove4DPacket(int totalTicks, short xTicks, short yTicks, short zTicks, short eTicks)
		{
			this.Data[1] = (byte)MessageID.MSGID_CONTROLLED_MOVE_4D;
			PackInt(2, totalTicks);
			PackShort(6, xTicks);
			PackShort(8, yTicks);
			PackShort(10, zTicks);
			PackShort(12, eTicks);
		}			
	}

	public class ControlledMove5DPacket : RepRapPacket
	{
		public ControlledMove5DPacket(int ticks,
				bool xdir, int dxi, int ddxi, int slopeX, int errorX,
				bool ydir, int dyi, int ddyi, int slopeY, int errorY,
				bool edir, int dei, int ddei, int slopeE, int errorE)
		{
			this.Data[1] = (byte)MessageID.MSGID_CONTROLLED_MOVE_5D;
			this.Data[2] = 0;
			if (xdir)
				this.Data[2] |= 1;
			if (ydir)
				this.Data[2] |= 2;
			if (edir)
				this.Data[2] |= 4;
			PackInt(3, ticks);
			PackInt(7, dxi);
			PackInt(11, ddxi);
			PackInt(15, slopeX);
			PackInt(19, errorX);
			PackInt(23, dyi);
			PackInt(27, ddyi);
			PackInt(31, slopeY);
			PackInt(35, errorY);
			PackInt(39, dei);
			PackInt(43, ddei);
			PackInt(47, slopeE);
			PackInt(51, errorE);
		}
	}

	public class DwellPacket : RepRapPacket
	{
		public DwellPacket(double dwellTime)
		{
			this.Data[1] = (byte)MessageID.MSGID_DWELL;
			PackShort(2, (int)dwellTime);
		}
	}

	public enum MessageID : byte
	{
		MSGID_REQUEST_STATUS		= 0x10,
		MSGID_REQUEST_TARGET_TEMP	= 0x11,

		MSGID_XSTEPPER				= 0x20,
		MSGID_YSTEPPER				= 0x21,
		MSGID_ZSTEPPER				= 0x22,
		MSGID_ESTEPPER				= 0x23,

		MSGID_EXTRUDER_HEATER		= 0x30,
		MSGID_HEATBED_HEATER		= 0x31,
		
		MSGID_SET_TARGET_TEMP		= 0x40,

		MSGID_GO_HOME				= 0x60,
		MSGID_CONTROLLED_MOVE_4D	= 0x61,
		MSGID_DWELL					= 0x62,
		MSGID_CONTROLLED_MOVE_5D	= 0x63,

		MSGID_STATUS				= 0x80,
		MSGID_TEMP_SETTING			= 0x81,
	}
}
