﻿//#define NO_HARDWARE		// used for debugging prints without an actual reprap connected
//#define NO_EXTRUSION		// used for debugging prints using XYZ motion only
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Diagnostics;
using System.Threading;
using System.Windows.Threading;

namespace Repic
{
	public class RepRapPrinter : BasePrinter, INotifyPropertyChanged
	{
		// needed for two-way binding
		public event PropertyChangedEventHandler PropertyChanged;
		public void OnPropertyChanged(PropertyChangedEventArgs e)
		{
			if (PropertyChanged != null)
				PropertyChanged(this, e);
		}

		public RepRapPrinter(Window parent)
			: base(parent)
		{
			// initialize the usb interface
			this.Usb = new Usb(parent);
		}

		protected override void Update()
		{
			if (this.Usb == null)
				return;
			base.Update();
			GetStatus();
		}		

		// sends a packet to the RepRap asking it to report it's status and sleeps until it receives the result
		private void GetStatus()
		{			
#if NO_HARDWARE
			this.Parent.Dispatcher.Invoke(DispatcherPriority.Normal, (ThreadStart)delegate()
			{
				this.XEndStop = true;
				this.YEndStop = false;
				this.ZEndStop = false;
				this.ExtruderADC = TempToADC(210);
				this.SlotsAvailable = 10;
			});

#else	// #if NO_HARDWARE

			this.SlotsAvailable = 0;

			// tell the RepRap to send its status
			SendUsbPacket(new RequestStatusPacket());

			// wait for it to respond (TO DO: this needs to be made a bit more stable)
			RepRapPacket status = this.Usb.WaitForPacket();

			// process the data
			if (status.Data[1] == (byte)MessageID.MSGID_STATUS)
			{
				this.ExtruderADC = TempToADC(200);
				this.Parent.Dispatcher.Invoke(DispatcherPriority.Normal, (ThreadStart)delegate()
				{
					this.XEndStop = (status.Data[2] != 0);
					this.YEndStop = (status.Data[3] != 0);
					this.ZEndStop = (status.Data[4] != 0);
					this.ExtruderADC = (ushort)((int)status.Data[5] + (int)status.Data[6] * 256);
					this.SlotsAvailable = status.Data[7];
				});
			}
#endif	// #if NO_HARDWARE
		}

		public override bool IsBusy
		{
			get
			{
				return (this.SlotsAvailable < 2) && (!base.IsBusy);
			}
		}

		public override void GoHome(PrintJob job, double? X, double? Y, double? Z)
		{
			base.GoHome(job, X, Y, Z);
			this.currentX = 0;
			this.currentY = 0;

			SendUsbPacket(new GoHomePacket(X, Y, Z));
			this.SlotsAvailable--;
		}

		double currentX = 0;
		double currentY = 0;

		public override void ControlledMove(PrintJob job, ExtrudeData extrudeData)
		{
			base.ControlledMove(job, extrudeData);
			double nextX = currentX + (extrudeData.End.X - extrudeData.Start.X);
			double nextY = currentY + (extrudeData.End.Y - extrudeData.Start.Y);
			extrudeData.Start.X = currentX;
			extrudeData.Start.Y = currentY;
			extrudeData.End.X = nextX;
			extrudeData.End.Y = nextY;
			currentX = nextX;
			currentY = nextY;
			//SendExtrusion4D(job, extrudeData);
			SendExtrusion5D(job, extrudeData);
			this.SlotsAvailable--;
		}

		public override void Dwell(PrintJob job, double dwellTime)
		{
			base.Dwell(job, dwellTime);
			SendUsbPacket(new DwellPacket(dwellTime));
			this.SlotsAvailable--;
		}

		public ConnectionStatus ConnectionStatus
		{
#if NO_HARDWARE
			get { return ConnectionStatus.Connected; }
#else
			get { return this.Usb.Status; }
#endif
		}

		public event StatusChangedDelegate OnStatusChanged
		{
			add
			{
				//this.Usb.OnStatusChanged += value;
			}
			remove
			{
				//this.Usb.OnStatusChanged -= value;
			}
		}

		
		private int SlotsAvailable = 0;

		#region Step Rate Settings

		const double INCHES_TO_MM = 2.54;
		const double EPSILON = 0.01;	// 1/10th resolution of the machine
		const double EPSILON_SQUARED = EPSILON * EPSILON;
		const double STEP_SIZE = 16;
		const double STEP_RATE = 200;			// for a full circle
		const double NUM_PULLEY_TEETH = 8;
		const double NUM_GEAR_TEETH = 21;		// number of teeth on the z axis gear
		const double GEAR_PITCH = 5;			// gear pitch in mm
		const double Z_AXIS_PITCH = 1.25;		// mm

		public static double X_STEPS_PER_MM = STEP_SIZE * STEP_RATE / GEAR_PITCH / NUM_PULLEY_TEETH;
		public static double Y_STEPS_PER_MM = STEP_SIZE * STEP_RATE / GEAR_PITCH / NUM_PULLEY_TEETH;
		public static double Z_STEPS_PER_MM = STEP_SIZE * STEP_RATE * NUM_GEAR_TEETH / NUM_PULLEY_TEETH / Z_AXIS_PITCH;
		//public const double E_STEPS_PER_MM = 17.6f;		// Adrian's geared extruder.
		//public const double E_STEPS_PER_MM = 22.0f;		// Adrian's geared extruder with crappy drive gear.
		public static double E_STEPS_PER_MM = 13.50f;		// Wade's extruder
		public const double TICKS_PER_MINUTE = 10000 * 60;	// PIC interrupt rate

		#endregion // Step Rate Settings

		#region Extruder Thermistor Control

		static double r0 = 100000;									// stated resistance, e.g. 10K
		static double r1 = 4700;
		static double r2 = 10000;
		static double t0 = 25.0 + 273.15;							// temperature at stated resistance, e.g. 25C
		static double beta = 4066;									// stated beta, e.g. 3500
		static double vadc = 5.0;									// ADC reference
		static double vcc = 5.0;									// supply voltage to potential divider
		static double vs = r1 * vcc / (r1 + r2);					// effective bias voltage
		static double rs = r1 * r2 / (r1 + r2);						// effective bias impedance
		static double k = r0 * Math.Exp(-beta / t0);				// constant part of calculation

		private int _ExtruderTemperature = 20;
		public int ExtruderTemperature
		{
			get { return this._ExtruderTemperature; }
			private set {
				this._ExtruderTemperature = value;
				OnPropertyChanged(new PropertyChangedEventArgs("ExtruderTemperature"));
			}
		}

		private int _ExtruderADC = 0xffff;
		private int ExtruderADC
		{
			get { return this._ExtruderADC; }
			set
			{
				this._ExtruderADC = value;

				// convert to a temperature and set the thermometer control
				double v = value * vadc / 1023;						// convert the 10 bit ADC value to a voltage
				double r = rs * v / (vs - v);						// resistance of thermistor
				double temp = (beta / Math.Log(r / k)) - 273.15;	// temperature
				this.ExtruderTemperature = (int)temp;				
			}
		}

		static ushort TempToADC(float t)
		{
			double r = r0 * Math.Exp(beta * (1 / (t + 273.15) - 1 / t0)); // resistance of the thermistor
			double v = vs * r / (rs + r);	// the voltage at the potential divider
			return (ushort)Math.Round(v / vadc * 1023, MidpointRounding.AwayFromZero);  // the ADC reading
		}

		static int ADCToTemp(int adc)
		{
			double v = adc * vadc / 1023;						// convert the 10 bit ADC value to a voltage
			double r = rs * v / (vs - v);						// resistance of thermistor
			double temp = (beta / Math.Log(r / k)) - 273.15;	// temperature
			return (int)Math.Round(temp, MidpointRounding.AwayFromZero);
		}

		#endregion Extruder Thermistor Control

		#region Stepper Manual Control

		private int _XStepper = 0;
		public int XStepper
		{
			get { return this._XStepper; }
			set
			{
				if (this._XStepper != value)
				{
					this._XStepper = value;
					SendUsbPacket(new XStepperPacket(this._XStepper));
				}
			}
		}

		private int _YStepper = 0;
		public int YStepper
		{
			get { return this._YStepper; }
			set
			{
				if (this._YStepper != value)
				{
					this._YStepper = value;
					SendUsbPacket(new YStepperPacket(this._YStepper));
				}
			}
		}

		private int _ZStepper = 0;
		public int ZStepper
		{
			get { return this._ZStepper; }
			set
			{
				if (this._ZStepper != value)
				{
					this._ZStepper = value;
					SendUsbPacket(new ZStepperPacket(this._ZStepper));
				}
			}
		}

		private int _EStepper = 0;
		public int EStepper
		{
			get { return this._EStepper; }
			set
			{
				if (this._EStepper != value)
				{
					this._EStepper = value;
#if !NO_EXTRUSION
					SendUsbPacket(new EStepperPacket(this._EStepper));
#endif
				}
			}
		}
		
		#endregion // Stepper Manual Control

		#region Extruder Heater Control

		private bool _ExtruderHeater = false;
		public bool ExtruderHeater
		{
			get { return this._ExtruderHeater; }
			set
			{
				if (this._ExtruderHeater != value)
				{
					this._ExtruderHeater = value;
					SendUsbPacket(new ExtruderHeaterPacket(this._ExtruderHeater));
				}
			}
		}

		private bool _HeatbedHeater = false;
		public bool HeatbedHeater
		{
			get { return this._HeatbedHeater; }
			set
			{
				if (this._HeatbedHeater != value)
				{
					this._HeatbedHeater = value;
					SendUsbPacket(new HeatbedHeaterPacket(this._HeatbedHeater));
				}
			}
		}

		#endregion Extruder Heater Control

		#region End Stop Control

		private bool _XEndStop = false;
		public bool XEndStop
		{
			get { return this._XEndStop; }
			set { this._XEndStop = value; OnPropertyChanged(new PropertyChangedEventArgs("XEndStop")); }
		}

		private bool _YEndStop = false;
		public bool YEndStop
		{
			get { return this._YEndStop; }
			set { this._YEndStop = value; OnPropertyChanged(new PropertyChangedEventArgs("YEndStop")); }
		}

		private bool _ZEndStop = false;
		public bool ZEndStop
		{
			get { return this._ZEndStop; }
			set { this._ZEndStop = value; OnPropertyChanged(new PropertyChangedEventArgs("ZEndStop")); }
		}

		#endregion End Stop Control
		
		#region USB Communication

		// this member handles low-level packet communication with the target hardware
		volatile private Usb Usb = null;

		private void SendUsbPacket(RepRapPacket msg)
		{
			if (this.Usb == null)
				return;
			if (this.Usb.Status != ConnectionStatus.Connected)
				return;
			try
			{
				this.Usb.SendPacket(msg);
			}
			catch (Exception)
			{
			}
		}

		#endregion USB Communication

		public void SetTargetTemp(int temp)
		{
			SendUsbPacket(new SetTargetTempPacket(temp, TempToADC(temp)));
		}

		public int GetTargetTemp()
		{
			// tell the RepRap to send its status
			SendUsbPacket(new RequestTargetTempPacket());

			// wait for it to respond (TO DO: this needs to be made a bit more stable)
			RepRapPacket packet = this.Usb.WaitForPacket();

			// process the data
			if (packet.Data[1] == (byte)MessageID.MSGID_TEMP_SETTING)
				return packet.Data[2];

			return 220;	// default
		}

		// based on cartesian_dda::set_target in the official firmware
		public void SendExtrusion4D(PrintJob job, ExtrudeData extrudeData)
		{
#if NO_EXTRUSION
			extrudeData.Start.E = extrudeData.End.E = 0;
#endif

			// calculate our units, adjusting for imperial scale if we need to
			double scale = (job.CurrentUnitType == UnitType.Imperial) ? INCHES_TO_MM : 1.0;
			Vector5D units = new Vector5D(
				X_STEPS_PER_MM * scale, Y_STEPS_PER_MM * scale, Z_STEPS_PER_MM * scale, E_STEPS_PER_MM * scale, 1.0);

			// calculate distance to move along each axis
			Vector5D delta = extrudeData.End - extrudeData.Start;

			// calculate how far we're going to move the extrude head
			Vector5D squares = delta * delta;
			double distance = squares.X + squares.Y + squares.Z;

			// if it's 0 then the only thing changing is e
			if (distance < EPSILON_SQUARED)
				distance = squares.E * squares.E;

			// if it's still zero then the only thing changing is the feed rate
			if (distance < EPSILON_SQUARED)
				return;

			// calculate the total distance
			distance = Math.Sqrt(distance);

			// calculate the time (in interrupt ticks) based on the feedrate
			if (extrudeData.End.F <= 0)
				return;
			int totalTicks = (int)(60 * 10000 * distance / extrudeData.End.F);
			if (totalTicks <= 0)
				return;

			// calculate our current, target and delta steps
			Vector5D currentSteps = ToSteps(units, extrudeData.Start);
			Vector5D targetSteps = ToSteps(units, extrudeData.End);
			Vector5D deltaSteps = targetSteps - currentSteps;
			short xTicks = (short)deltaSteps.X;
			short yTicks = (short)deltaSteps.Y;
			short zTicks = (short)deltaSteps.Z;
			short eTicks = (short)deltaSteps.E;
			if ((xTicks == 0) && (yTicks == 0) && (zTicks == 0) && (eTicks == 0))
				return;

			// make sure we don't try to print too fast - todo: increase the ISR freq in the PIC?
			Debug.Assert(totalTicks >= xTicks << 1);
			Debug.Assert(totalTicks >= yTicks << 1);
			//Debug.Assert(totalTicks >= zTicks << 1);
			//Debug.Assert(totalTicks >= eTicks << 1);
			totalTicks = Math.Max(totalTicks, xTicks << 1);
			totalTicks = Math.Max(totalTicks, yTicks << 1);
			totalTicks = Math.Max(totalTicks, zTicks << 1);
			totalTicks = Math.Max(totalTicks, eTicks << 1);

			// send the data to the repap
			SendUsbPacket(new ControlledMove4DPacket(totalTicks, xTicks, yTicks, zTicks, eTicks));
		}

		Vector5D ToSteps(Vector5D units, Vector5D v)
		{
			v *= units;
			v.X = (int)v.X;
			v.Y = (int)v.Y;
			v.Z = (int)v.Z;
			v.E = (int)v.E;
			v.F = (int)v.F;
			return v;
		}

		/*
		 * FiveD Implementation Notes
		 * ==========================
 
		 * The math behind FiveD extrusion is the same as that for accelerated motion, where 'feed rate' is
		 * analogous to 'velocity'. From newtonian physics we have the basic equation of accelerated motion:

		 * 1) v=u+a*t

		 * where	v = final velocity (end feedrate)
		 * 		u = initial velocity (start feedrate)
		 * 		a = acceleration
		 * 		t = time

		 * By integrating with respect to t we can calculate total movement/extrusion distance over a given time
		 * period:

		 * 2) s=u*t + 0.5*a*t^2

		 * This is the equation we will use at run-time to determine how far the stepper motors should move. First,
		 * we must calculate 'a', which we can do if we know the total time for the entire extrusion. By rearranging
		 * 1) to solve for 'a' and substituting into 2) we arrive at equation 3) which allows us to calculate total
		 * extrude time 't' given start and end feed rates and the total distance 's', which is how extrusion is
		 * specified in gcode:

		 * 3) t = 2 * s / (u + v)

		 * (Note that conceptually this makes sense, it's simply distance divided by the average feed rate).

		 * In the case of RepRap our feed rates are specified in mm/minute, so the calculated 't' value is converted
		 * from minutes to the interrupt tick-rate on the micro-controller (i.e. 10000/second in the case of Repic).
		 * This value is then rounded to the nearest integer and equation 1) is used to calculate 'a'. The rounding
		 * of 't' to the nearest integer often results in a small error being introduced to 'a' which in turn can
		 * cause the target feed rate value 'v' to be different at the end of the extrude. This error is well within
		 * the tolerences of RepRap though (typically < 0.01%) and the rounding ensures that the final iterated values
		 * of movement and extrusion equal the target 's' values exactly, which is more important.

		 * To do the interpolation we convert to stepper values by multiplying distance by the step rate of the
		 * stepper motor in question (typically ~80 for the X and Y axis on Mendel). Each step requires two
		 * interrupts, one to set the driver signal to 1 and another to set it back to 0, so we double this value
		 * and then scale equation 2) by the result in order to calculate the number of step signals generated for
		 * any given t. We do not, however, need to fully evaluate equation 2) for each iteration in the main loop
		 * because s(t+1) can be calculated as a function of s(t):

		 * 	s(t+1)	= u * (t+1) + 0.5 * a * (t+1) * (t+1)
		 * 		 * = u * t + u + 0.5 * a * (t * t + 2 * t + 1)
		 * 		 * = (u * t + 0.5 * a * t * t) + a * t + 0.5 * a + u 
		 * 		 * = s(t) + (a * t) + (0.5 * a + u);
 
		 * The algorithm can therefore be implemented iteratively with floating-point variables using addition only
		 * in the inner loop:
 
		 * 		dNumSteps = 0.5 * a + u
		 * 		ddNumSteps = a
		 * 		numSteps = 0
		 * 		for (i=0; i<t; i++, numSteps += dNumSteps, dNumSteps += ddNumSteps)
		 * 		 * ... use numSteps here...

		 * In order to implement this code on a 16-bit PIC micro-controller the values are stored in 32-bit integer
		 * variables as 1:31 fixed-point values. Only 1 integer bit is needed for dNumSteps and ddNumSteps because
		 * RepRap cannot generate more than 1 step per interrupt tick, i.e. both values are always <= 1. We do lose
		 * all integer bits in numSteps apart from the LSB, but that's ok because only the LSB is needed to determine
		 * when the signal sent to the stepper driver should be toggled.

		 * Unfortunately, even 31 bits of precision can cause enough loss of precision over the course of an extrude
		 * to result in the final value of numSteps being slightly off-target. The total amount of this accumulated
		 * error at any given moment can be calculated exactly and is equal to the fractional component of ddNumSteps
		 * lost after conversion to 1:31 multiplied by the number of iterations:

		 * 	ddNumStepsScaled = ddNumSteps * (1 << 31);		// scale up for 1:31 fixed-point
		 * 	ddNumStepsInt = (int)ddNumStepsScaled;		 * // and store in a 32-bit integer
		 * 	errorTerm = (ddNumStepsScaled - ddNumStepsInt);	// the difference is the error for each iteration
		 * 	totalError = errorTerm * t;		 * 		 * // total error for the entire extrude

		 *  We can adjust for this by calculating this total error term iteratively using Brensenham's line algorithm
		 * and adding the integer component of the result to numSteps during the inner loop in addition to dNumSteps.
		 * This results in perfect FiveD integrated extrusion with zero accumulated error over the ranges of distance,
		 * feed rate and acceleration typically used in RepRap.
		 */

		private void SendExtrusion5D(PrintJob job, ExtrudeData extrudeData)
		{
			if ((extrudeData.Start.F == 0) && (extrudeData.End.F == 0))
				return; // nothing to do

#if NO_EXTRUSION
			extrudeData.Start.E = extrudeData.End.E = 0;
#endif

			// if we have a z component then treat it as a 4D move
			if (extrudeData.Start.Z != extrudeData.End.Z)
			{
				SendExtrusion4D(job, new ExtrudeData(
					new Vector5D(0, 0, extrudeData.Start.Z, 0, extrudeData.Start.F),
					new Vector5D(0, 0, extrudeData.End.Z, 0, extrudeData.End.F)
					));
			}

			// clamp the XYE feedrates to the limit of the machine
			if (extrudeData.Start.X != extrudeData.End.X)
			{
				double maxFeedrateX = TICKS_PER_MINUTE / X_STEPS_PER_MM / 2;
				extrudeData.Start.F = Math.Min(extrudeData.Start.F, maxFeedrateX);
				extrudeData.End.F = Math.Min(extrudeData.End.F, maxFeedrateX);
			}
			if (extrudeData.Start.Y != extrudeData.End.Y)
			{
				double maxFeedrateY = TICKS_PER_MINUTE / Y_STEPS_PER_MM / 2;
				extrudeData.Start.F = Math.Min(extrudeData.Start.F, maxFeedrateY);
				extrudeData.End.F = Math.Min(extrudeData.End.F, maxFeedrateY);
			}

			// calculate our units, adjusting for imperial scale if we need to
			double scale = (job.CurrentUnitType == UnitType.Imperial) ? INCHES_TO_MM : 1.0;
			Vector5D stepRates = new Vector5D(
				X_STEPS_PER_MM * scale, Y_STEPS_PER_MM * scale, Z_STEPS_PER_MM * scale, E_STEPS_PER_MM * scale,
				TICKS_PER_MINUTE);

			// calculate how far we're going to move along each axis
			Vector5D delta = extrudeData.End - extrudeData.Start;
			bool xdir = delta.X < 0;
			bool ydir = delta.Y < 0;
			bool edir = delta.E < 0;
			delta.X = Math.Abs(delta.X);
			delta.Y = Math.Abs(delta.Y);
			delta.E = Math.Abs(delta.E);

			// calculate how far we're going to move the extrude head
			double distance = Math.Max(delta.X, delta.Y);

			// if it's 0 then the only thing changing is e
			if (distance < EPSILON)
				distance = delta.E;

			// if it's still zero then the only thing changing is the feed rate
			if (distance < EPSILON)
			    return;

			// solve for t
			double fi = extrudeData.Start.F;
			double fe = extrudeData.End.F;
			double t = 2 * distance / (fi + fe);

			// convert t from min to interrupts (10,000 per second) and round to nearest integer
			int ticks = (int)Math.Round(t * stepRates.F, MidpointRounding.AwayFromZero);

			// calculate a modified value of a
			double ascale = stepRates.F / ticks;
			double a = 2.0 * (distance * ascale - fi) * ascale;

			delta.X = Math.Round(delta.X * stepRates.X * 2, MidpointRounding.AwayFromZero);
			delta.Y = Math.Round(delta.Y * stepRates.Y * 2, MidpointRounding.AwayFromZero);
			delta.E = Math.Round(delta.E * stepRates.E * 2, MidpointRounding.AwayFromZero);

			double fpscale = 1U << 31;
			int fpscalei = (int)(0.5 * fpscale);
			double tscale = t / ticks;

			double ax = a * delta.X * tscale * tscale / distance,
					fix = fi * delta.X * tscale / distance,
					dx = (ax * 0.5 + fix) * fpscale,
					ddx = ax * fpscale;
			int dxi = (int)Math.Floor(dx),
				ddxi = (int)Math.Floor(ddx),
				slopeX = (int)(ticks * (ddx - ddxi)),
				errorX = (int)((dx - dxi - 1) * ticks);

			double ay = a * delta.Y * tscale * tscale / distance,
					fiy = fi * delta.Y * tscale / distance,
					dy = (ay * 0.5 + fiy) * fpscale,
					ddy = ay * fpscale;
			int dyi = (int)Math.Floor(dy),
				ddyi = (int)Math.Floor(ddy),
				slopeY = (int)(ticks * (ddy - ddyi)),
				errorY = (int)((dy - dyi - 1) * ticks);

			double ae = a * delta.E * tscale * tscale / distance,
					fie = fi * delta.E * tscale / distance,
					de = (ae * 0.5 + fie) * fpscale,
					dde = ae * fpscale;
			int dei = (int)Math.Floor(de),
				ddei = (int)Math.Floor(dde),
				slopeE = (int)(ticks * (dde - ddei)),
				errorE = (int)((de - dei - 1) * ticks);

			SendUsbPacket(new ControlledMove5DPacket(ticks,
				xdir, dxi, ddxi, slopeX, errorX,
				ydir, dyi, ddyi, slopeY, errorY,
				edir, dei, ddei, slopeE, errorE));
		}

		
	}
}
