﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Input;
using System.Windows.Media.Imaging;
using System.Windows.Shapes;
using System.Globalization;

namespace Repic
{
	public class Thermometer : FrameworkElement
	{
		public static readonly DependencyProperty MinTempProperty;
		public static readonly DependencyProperty MaxTempProperty;
		public static readonly DependencyProperty TargetTempProperty;
		public static readonly DependencyProperty TempStepProperty;
		public static readonly DependencyProperty GradStepProperty;
		public static readonly DependencyProperty CurrentTempProperty;

		static Thermometer()
		{
			FrameworkPropertyMetadata minMeta = new FrameworkPropertyMetadata(
				(double)20, FrameworkPropertyMetadataOptions.AffectsRender);
			MinTempProperty = DependencyProperty.Register("MinTemp",
				typeof(double), typeof(FrameworkElement), minMeta, null);

			FrameworkPropertyMetadata maxMeta = new FrameworkPropertyMetadata(
				(double)300, FrameworkPropertyMetadataOptions.AffectsRender);
			MaxTempProperty = DependencyProperty.Register("MaxTemp",
				typeof(double), typeof(FrameworkElement), maxMeta, null);

			FrameworkPropertyMetadata tempStepMeta = new FrameworkPropertyMetadata(
				(double)10, FrameworkPropertyMetadataOptions.AffectsRender);
			TempStepProperty = DependencyProperty.Register("TempStep",
				typeof(double), typeof(FrameworkElement), tempStepMeta, null);

			FrameworkPropertyMetadata gradStepMeta = new FrameworkPropertyMetadata(
				(double)5, FrameworkPropertyMetadataOptions.AffectsRender);
			GradStepProperty = DependencyProperty.Register("GradStep",
				typeof(double), typeof(FrameworkElement), gradStepMeta, null);

			FrameworkPropertyMetadata targetMeta = new FrameworkPropertyMetadata(
				(double)250, FrameworkPropertyMetadataOptions.AffectsRender);
			TargetTempProperty = DependencyProperty.Register("TargetTemp",
				typeof(double), typeof(FrameworkElement), targetMeta, null);

			FrameworkPropertyMetadata currentMeta = new FrameworkPropertyMetadata(
				(double)5, FrameworkPropertyMetadataOptions.AffectsRender);
			CurrentTempProperty = DependencyProperty.Register("CurrentTemp",
				typeof(double), typeof(FrameworkElement), currentMeta, null);
		}

		private bool IsTargetTempValid(object value)
		{
			return ((double)value >= this.MinTemp) && ((double)value <= this.MaxTemp);
		}

		public double MinTemp
		{
			set { SetValue(MinTempProperty, value); }
			get { return (double)GetValue(MinTempProperty); }
		}

		public double MaxTemp
		{
			set { SetValue(MaxTempProperty, value); }
			get { return (double)GetValue(MaxTempProperty); }
		}

		public double TargetTemp
		{
			set { SetValue(TargetTempProperty, value); }
			get { return (double)GetValue(TargetTempProperty); }
		}

		public double TempStep
		{
			set { SetValue(TempStepProperty, value); }
			get { return (double)GetValue(TempStepProperty); }
		}

		public double GradStep
		{
			set { SetValue(GradStepProperty, value); }
			get { return (double)GetValue(GradStepProperty); }
		}

		public double CurrentTemp
		{
			set { SetValue(CurrentTempProperty, value); }
			get { return (double)GetValue(CurrentTempProperty); }
		}

		protected double TempToLine(double degrees)
		{
			double result = 440 * (degrees - this.MinTemp) / (this.MaxTemp - this.MinTemp);
			if (result < 0)
				result = 0;
			if (result > 440)
				result = 440;
			return 480 - result;
		}

		protected override void OnRender(DrawingContext dc)
		{
			base.OnRender(dc);
			Pen pen = new Pen(Brushes.Blue, 6);

			Rect bowlRect = new Rect(100, 575, 50, 50);

			// draw the bulb and stem
			LinearGradientBrush stemBrush = new LinearGradientBrush();			
			stemBrush.StartPoint = new Point(0, 0);
			stemBrush.EndPoint = new Point(1, 0);
			stemBrush.GradientStops.Add(new GradientStop(Colors.Red, 0));
			stemBrush.GradientStops.Add(new GradientStop(Colors.LightPink, 0.3));
			stemBrush.GradientStops.Add(new GradientStop(Colors.Red, 1));
			RadialGradientBrush bulbBrush = new RadialGradientBrush();
			bulbBrush.GradientOrigin = new Point(0.25, 0.25);
			bulbBrush.GradientStops.Add(new GradientStop(Colors.LightPink, 0));
			bulbBrush.GradientStops.Add(new GradientStop(Colors.Red, 1));
			double tempy = TempToLine(CurrentTemp);
			dc.DrawRoundedRectangle(null, pen, new Rect(75, 25, 50, 550), 25, 25);
			dc.DrawEllipse(null, pen, bowlRect.Location, bowlRect.Width, bowlRect.Height);
			dc.DrawRoundedRectangle(Brushes.SkyBlue, null, new Rect(75, 25, 50, 550), 25, 25);
			dc.DrawRoundedRectangle(stemBrush, null, new Rect(75, tempy, 50, 550 - tempy), 10, 10);
			dc.DrawEllipse(bulbBrush, null, new Point(100, 575), 50, 50);

			// draw the graduations
			Pen minorGradPen = new Pen(Brushes.Black, 2);
			Pen majorGradPen = new Pen(Brushes.Black, 2);
			Typeface typeface = new Typeface("Arial");
			int gradFontHeight = 22;
			for (double degrees = this.MinTemp; degrees <= this.MaxTemp; degrees += this.GradStep)
			{
				double y = TempToLine(degrees);
				if (degrees % this.TempStep == 0)
				{
					FormattedText text = new FormattedText(Convert.ToString(degrees), CultureInfo.CurrentCulture,
						FlowDirection.RightToLeft, typeface, gradFontHeight, Brushes.Black);
					dc.DrawText(text, new Point(70, y - gradFontHeight/2));
					dc.DrawLine(majorGradPen, new Point(90, y), new Point(110, y));
				}
				else
				{
					dc.DrawLine(minorGradPen, new Point(95, y), new Point(105, y));
				}
			}

			// draw the cut-off points
			PathGeometry path = new PathGeometry();
			path.FillRule = FillRule.Nonzero;

			// draw the target temp reticle
			double targetY = TempToLine(TargetTemp);
			PathFigure targetFigure = new PathFigure();
			targetFigure.IsClosed = true;
			targetFigure.StartPoint = new Point(130, targetY);
			targetFigure.Segments.Add(new LineSegment(new Point(140, targetY - 10), true));
			targetFigure.Segments.Add(new LineSegment(new Point(140, targetY + 10), true));
			path.Figures.Add(targetFigure);
			dc.DrawGeometry(Brushes.Black, null, path);

			// draw the current temp in the bowl
			FormattedText ft = new FormattedText(Convert.ToString(CurrentTemp) + "°C", CultureInfo.CurrentCulture,
				FlowDirection.LeftToRight, typeface, gradFontHeight, Brushes.Black);
			dc.DrawText(ft, new Point(bowlRect.Left - ft.Width / 2, bowlRect.Top - ft.Height / 2));
		}

	}
}
