﻿//#define PCB
//#define CONSTANT_EXTRUSION_4D
//#define CONSTANT_EXTRUSION_5D

using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Media.Media3D;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.ComponentModel;
using System.Threading;
using System.Windows.Threading;
using System.Runtime;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;

namespace Repic
{
	public class VirtualRepRap : BasePrinter
	{
		private Point3D PieceOffset = new Point3D(14.3, -2.23413461685181, 3.67);
		List<GeometryModel3D> CurrentChildModels = new List<GeometryModel3D>();

		DiffuseMaterial MeshMaterial = new DiffuseMaterial(Brushes.CornflowerBlue);
		GeometryModel3D CurrentModel;

		// queued list of extrusions ready for sending to the hardware
		Queue<ExtrudeData> Extrudes = new Queue<ExtrudeData>();
		//List<ExtrudeData> AllExtrudes = new List<ExtrudeData>();

		object RenderLock = new object();

		public VirtualRepRap(Window parent)
			: base(parent)
		{
			this.Parent = parent;
		}

		// idle processing worker thread, sits in a loop using spare background cycles to do rendering to the main window
		protected override void Update()
		{
			base.Update();
			this.Parent.Dispatcher.Invoke(DispatcherPriority.Normal, (ThreadStart)delegate()
			{
				lock (RenderLock)
					RenderLoop();
			});

			// Sleep for a bit so we don't tie up 100% of the CPU
			//Thread.Sleep(1000 / 200);
		}

		void RenderLoop()
		{
			// draw any extrusions
			lock (this.Extrudes)
				while (this.Extrudes.Count > 0)
					AddExtrusion(this.Extrudes.Dequeue());
		}		

		public override void GoHome(PrintJob job, double? X, double? Y, double? Z)
		{
			base.GoHome(job, X, Y, Z);
		}

		public override void ControlledMove(PrintJob job, ExtrudeData extrudeData)
		{
			base.ControlledMove(job, extrudeData);
			lock (this.Extrudes)
				this.Extrudes.Enqueue(extrudeData);
		}

		public override void Dwell(PrintJob job, double dwellTime)
		{
			base.Dwell(job, dwellTime);
		}

		private void AddExtrusion(ExtrudeData e)
		{
#if DEBUG
			//this.AllExtrudes.Add(e);
			FiveD.Test(e, new Vector5D(
				RepRapPrinter.X_STEPS_PER_MM,
				RepRapPrinter.Y_STEPS_PER_MM,
				RepRapPrinter.Z_STEPS_PER_MM,
				RepRapPrinter.E_STEPS_PER_MM,
				RepRapPrinter.TICKS_PER_MINUTE
				));
#endif

			if ((e.Start.X == e.End.X) && (e.Start.Y == e.End.Y) && (e.Start.Z == e.End.Z))
				return;
#if PCB
			if ((e.Start.Z > 0) || (e.End.Z > 0))
				return;
#else
			if (e.Start.E >= e.End.E)
				return;
#endif

			double dx = e.End.X - e.Start.X;
			double dy = e.End.Y - e.Start.Y;
			double dz = e.End.Z - e.Start.Z;
			double de = e.End.E - e.Start.E;
			double distance = Math.Sqrt(dx * dx + dy * dy + dz * dz);
#if CONSTANT_EXTRUSION_4D
			double widthA = 30.0 / e.End.F;
			double widthB = 30.0 / e.End.F;
#elif CONSTANT_EXTRUSION_5D
			double widthA = 30.0 / e.Start.F;
			double widthB = 30.0 / e.End.F;			
#elif PCB
			double widthA = 0.005;
			double widthB = 0.005;
#else		// correct 5D extrusion
			double scale = (0.0125 / 13.50f); // magic scaling number for rendering...don't ask...
			double widthA = scale * RepRapPrinter.E_STEPS_PER_MM * de / distance;	// ideally de / distance == 1
			double widthB = scale * RepRapPrinter.E_STEPS_PER_MM * de / distance;

#endif

			double s = 0.0392;		// geometry render scale
			Point3D p1 = new Point3D(e.Start.X * s + PieceOffset.X, e.Start.Y * s + PieceOffset.Y, e.Start.Z * s + PieceOffset.Z);
			Point3D p2 = new Point3D(e.End.X * s + PieceOffset.X, e.End.Y * s + PieceOffset.Y, e.End.Z * s + PieceOffset.Z);
			Vector3D up = new Vector3D(0, 0, 1);
			Vector3D y = p2 - p1;
			y.Normalize();
			Vector3D x = Vector3D.CrossProduct(y, up);
			x.Normalize();
			Vector3D z = Vector3D.CrossProduct(x, y);
			z.Normalize();
			Point3D a1 = p1 + widthA * z;
			Point3D a2 = p1 + widthA * x;
			Point3D a3 = p1 - widthA * x;
			Point3D b1 = p2 + widthB * z;
			Point3D b2 = p2 + widthB * x;
			Point3D b3 = p2 - widthB * x;
			AddPrim(a1, a2, b1, b2, z, x, z, x);
			AddPrim(a3, a1, b3, b1, -x, z, -x, z);
		}

		private void AddPrim(Point3D p0, Point3D p1, Point3D p2, Point3D p3,
							 Vector3D n0, Vector3D n1, Vector3D n2, Vector3D n3)
		{
			MeshGeometry3D mesh = this.CurrentModel.Geometry as MeshGeometry3D;						
			int index = mesh.Positions.Count;

			mesh.Positions.Add(p0);
			mesh.Positions.Add(p1);
			mesh.Positions.Add(p2);
			mesh.Positions.Add(p3);
			mesh.Normals.Add(n0);
			mesh.Normals.Add(n1);
			mesh.Normals.Add(n2);
			mesh.Normals.Add(n3);
			mesh.TriangleIndices.Add(index + 0);
			mesh.TriangleIndices.Add(index + 1);
			mesh.TriangleIndices.Add(index + 2);
			mesh.TriangleIndices.Add(index + 1);
			mesh.TriangleIndices.Add(index + 3);
			mesh.TriangleIndices.Add(index + 2);

			// don't let the mesh get too large, otherwise we'll be spending all our time rebuilding geometry.
			if (mesh.Positions.Count > 1000)
				CreateNewModel();
		}

		public override void Reset()
		{
			//Stream stream = File.Open("extrudes.dat", FileMode.Create);
			//BinaryFormatter bformatter = new BinaryFormatter();
			//bformatter.Serialize(stream, this.Extrudes);
			//stream.Close();
			//this.AllExtrudes.Clear();

			lock (RenderLock)
			{
				Model3DGroup group = this.Parent.Resources["MendelGroup"] as Model3DGroup;
				foreach (GeometryModel3D child in this.CurrentChildModels)
					group.Children.Remove(child);
				this.CurrentChildModels.Clear();
				CreateNewModel();
				this.Extrudes.Clear();
			}
		}

		private void CreateNewModel()
		{
			this.CurrentModel = new GeometryModel3D();
			this.CurrentModel.Material = this.MeshMaterial;
			MeshGeometry3D mesh = new MeshGeometry3D();
			this.CurrentModel.Geometry = mesh;
			mesh.Positions = new Point3DCollection();
			mesh.Normals = new Vector3DCollection();
			mesh.TriangleIndices = new Int32Collection();

			Model3DGroup group = this.Parent.Resources["MendelGroup"] as Model3DGroup;
			group.Children.Add(this.CurrentModel);
			this.CurrentChildModels.Add(this.CurrentModel);
		}

		/* test code used to generate animation files, too slow to be of practical use at this stage
		int frameNum = 0;

		private void SaveFrame()
		{
			// save to bitmaps
			RenderTargetBitmap bitmap = new RenderTargetBitmap(
				(int)this.mainViewport.ActualWidth, (int)this.mainViewport.ActualHeight,
				96, 96, PixelFormats.Pbgra32);
			bitmap.Render(this.mainViewport);
			PngBitmapEncoder image = new PngBitmapEncoder();
			image.Frames.Add(BitmapFrame.Create(bitmap));
			string filename = String.Format(@"Frames\frame_{0}.png", frameNum++);
			using (Stream fs = File.Create(filename))
				image.Save(fs);
		}
		*/
		
	}
}
